import io
import os
import re
import sys
import datetime
import json
import shutil
import tempfile
import pathlib
import logging
import platform
import signal
from collections import OrderedDict
from deriva.core import ErmrestCatalog, HatracStore, HatracJobAborted, HatracJobPaused, \
    HatracJobTimeout, urlquote, urlparse, stob, format_exception, get_credential, read_config, write_config, \
    copy_config, resource_path, make_dirs, lock_file, DEFAULT_CHUNK_SIZE, __version__ as VERSION
from deriva.core import DEFAULT_SESSION_CONFIG, DEFAULT_CREDENTIAL_FILE
from deriva.core.utils import hash_utils as hu, mime_utils as mu, version_utils as vu
from deriva.transfer.upload import *
from deriva.transfer.upload.processors import find_processor
from deriva.transfer.upload.processors.base_processor import *
from deriva.transfer.upload.processors.archive_processor import ArchiveProcessor

try:
    from os import scandir, walk
except ImportError:
    from scandir import scandir, walk

logger = logging.getLogger(__name__)


class Enum(tuple):
    __getattr__ = tuple.index


UploadState = Enum(["Success", "Failed", "Pending", "Running", "Paused", "Aborted", "Cancelled", "Timeout"])
UploadMetadataReservedKeyNames = [
    "URI", "file_name", "file_ext", "file_size", "base_path", "base_name", "content-disposition", "md5", "sha256",
    "md5_base64", "sha256_base64", "schema", "table", "target_table", "_upload_year_", "_upload_month_", "_upload_day_",
    "_upload_time_", "_identity_id", "_identity_display_name", "_identity_full_name", "_identity_email"]

DefaultConfig = {
  "version_compatibility": [[">=%s" % VERSION]],
  "version_update_url": "https://github.com/informatics-isi-edu/deriva-py/releases",
  "asset_mappings": [
    {
      "asset_type": "table",
      "default_columns": ["RID", "RCB", "RMB", "RCT", "RMT"],
      "file_pattern": "^((?!/assets/).)*/records/(?P<schema>.+?)/(?P<table>.+?)[.]",
      "ext_pattern": "^.*[.](?P<file_ext>json|csv)$"
    }
  ]
}


class FileUploadState:
    def __init__(self, state=UploadState.Pending, status="Pending", result=None):
        self.state = state
        self.status = status
        self.result = result

    def asdict(self):
        return OrderedDict({
            "State": self.state,
            "Status": self.status,
            "Result": self.result
        })


class UploadEntry(object):
    def __init__(self, asset_group, asset_mapping, groupdict, path):
        self.asset_group = asset_group
        self.asset_mapping = asset_mapping
        self.groupdict = groupdict
        self.path = path


class DerivaUpload(object):
    """
    Base class for upload tasks. Encapsulates a catalog instance and a hatrac store instance and provides some common
    and reusable functions.

    This class is not intended to be instantiated directly, but rather extended by a specific implementation.
    """

    DefaultConfigFileName = "config.json"
    DefaultServerListFileName = "servers.json"
    DefaultTransferStateBaseName = ".deriva-upload-state"
    DefaultTransferStateFileName = "%s-%s.json"

    def __init__(self, config_file=None, credential_file=None, server=None, dcctx_cid=None):
        self.server_url = None
        self.catalog = None
        self.catalog_model = None
        self.store = None
        self.config = None
        self.credentials = None
        self.asset_mappings = None
        self.transfer_state = dict()
        self.transfer_state_fh = None
        self.transfer_state_locks = dict()
        self.cancelled = False
        self.metadata = dict()
        self.catalog_metadata = {"table_metadata": {}}
        self.processor_output = dict()
        self.identity = dict()
        self.file_list = OrderedDict()
        self.file_status = OrderedDict()
        self.skipped_files = set()
        self.override_config_file = config_file
        self.override_credential_file = credential_file
        self.server = self.getDefaultServer() if not server else server
        self.dcctx_cid = dcctx_cid if dcctx_cid else self.__class__.__name__
        signal.signal(signal.SIGINT, self.interrupt_handler)
        self.initialize()

    def __del__(self):
        self.cleanupTransferState()

    def interrupt_handler(self, signum, frame):
        logger.info("Caught interrupt signal.")
        self.cancel()

    def initialize(self, cleanup=False):
        info = "%s v%s [Python %s, %s]" % (
            self.__class__.__name__, vu.get_installed_version(VERSION),
            platform.python_version(), platform.platform(aliased=True))
        logger.info("Initializing uploader: %s" % info)

        # cleanup invalidates the current configuration and credentials in addition to clearing internal state
        if cleanup:
            self.cleanup()
        # reset just clears the internal state
        else:
            self.reset()

        if not self.server:
            logger.warning("A server was not specified and an internal default has not been set.")
            return

        # server variable initialization
        protocol = self.server.get('protocol', 'https')
        host = self.server.get('host', '')
        self.server_url = protocol + "://" + host
        catalog_id = self.server.get("catalog_id", "1")
        session_config = self.server.get('session', DEFAULT_SESSION_CONFIG.copy())
        # default credential initialization
        self.credentials = get_credential(host, self.override_credential_file or DEFAULT_CREDENTIAL_FILE)

        # catalog and file store initialization
        if self.catalog:
            del self.catalog
        self.catalog = ErmrestCatalog(protocol, host, catalog_id, self.credentials, session_config=session_config)
        if self.store:
            del self.store
        self.store = HatracStore(protocol, host, self.credentials, session_config=session_config)

        # determine identity
        if self.credentials:
            attributes = self.catalog.get_authn_session().json()
            self.identity = attributes.get("client", self.identity)

        # init dcctx cid to a default
        self.set_dcctx_cid(self.dcctx_cid)

        """
         Configuration initialization - this is a bit complex because we allow for:
             1. Run-time overriding of the config file location.
             2. Sub-classes of this class to bundle their own default configuration files in an arbitrary location.
             3. The updating of already deployed configuration files if bundled internal defaults are newer.             
        """
        if self.override_config_file and not os.path.isfile(self.override_config_file):
            raise DerivaUploadConfigurationError(
                "The configuration file %s could not be found." % self.override_config_file)

        config_file = self.override_config_file if self.override_config_file else None
        # 1. If we don't already have a valid (i.e., overridden) path to a config file...
        if not (config_file and os.path.isfile(config_file)):
            # 2. Get the currently deployed config file path, which could possibly be overridden by subclass
            config_file = self.getDeployedConfigFilePath()
            # 3. If the deployed default path is not valid, OR, it is valid AND is older than the bundled default
            if (not (config_file and os.path.isfile(config_file))
                    or self.isFileNewer(self.getDefaultConfigFilePath(), self.getDeployedConfigFilePath())):
                # 4. If we can locate a bundled default config file,
                if os.path.isfile(self.getDefaultConfigFilePath()):
                    # 4.1 Copy the bundled default config file to the deployment-specific config path
                    copy_config(self.getDefaultConfigFilePath(), config_file)
                else:
                    # 4.2 Otherwise, fallback to writing a failsafe default based on internal hardcoded settings
                    write_config(config_file, DefaultConfig)
        # 5. Finally, read the resolved configuration file into a config object
        self._update_internal_config(read_config(config_file))

    def set_dcctx_cid(self, cid):
        assert cid, "A dcctx cid is required"
        if self.catalog:
            self.catalog.dcctx['cid'] = cid
        if self.store:
            self.store.dcctx['cid'] = cid

    def _update_internal_config(self, config):
        """This updates the internal state of the uploader based on the config.
        """
        self.config = config
        # uploader initialization from configuration
        self.asset_mappings = self.config.get('asset_mappings', [])
        mu.add_types(self.config.get('mime_overrides'))

    def cancel(self):
        self.cancelled = True

    def reset(self):
        self.metadata.clear()
        self.file_list.clear()
        self.file_status.clear()
        self.skipped_files.clear()
        self.cleanupTransferState()
        self.cancelled = False

    def cleanup(self):
        self.reset()
        self.config = None
        self.credentials = None
        self.catalog_model = None

    def setServer(self, server):
        cleanup = self.server != server
        self.server = server
        self.initialize(cleanup)

    def setCredentials(self, credentials):
        host = self.server['host']
        self.credentials = credentials
        self.catalog.set_credentials(self.credentials, host)
        self.store.set_credentials(self.credentials, host)
        attributes = self.catalog.get_authn_session().json()
        self.identity = attributes.get("client", self.identity)

    def setConfig(self, config_file):
        if not config_file:
            config = self.getUpdatedConfig()
            if config:
                write_config(self.getDeployedConfigFilePath(), config)
        else:
            self._update_internal_config(read_config(config_file))
        if not self.isVersionCompatible():
            raise RuntimeError("Upload version incompatibility detected",
                               "Current version: [%s], required version(s): %s." %
                               (self.getVersion(), self.getVersionCompatibility()))
    @classmethod
    def getDefaultServer(cls):
        servers = cls.getServers()
        for server in servers:
            lower = {k.lower(): v for k, v in server.items()}
            if lower.get("default", False):
                return server
        return servers[0] if len(servers) else {}

    @classmethod
    def getServers(cls):
        """
        This method must be implemented by subclasses.
        """
        raise NotImplementedError("This method must be implemented by a subclass.")

    @classmethod
    def getVersion(cls):
        """
        This method must be implemented by subclasses.
        """
        raise NotImplementedError("This method must be implemented by a subclass.")

    @classmethod
    def getConfigPath(cls):
        """
        This method must be implemented by subclasses.
        """
        raise NotImplementedError("This method must be implemented by a subclass.")

    @classmethod
    def getDeployedConfigPath(cls):
        return os.path.expanduser(os.path.normpath(cls.getConfigPath()))

    def getVersionCompatibility(self):
        return self.config.get("version_compatibility", list())

    def isVersionCompatible(self):
        compatibility = self.getVersionCompatibility()
        if len(compatibility) > 0:
            return vu.is_compatible(self.getVersion(), compatibility)
        else:
            return True

    @classmethod
    def getFileDisplayName(cls, file_path, asset_mapping=None):
        return os.path.basename(file_path)

    @staticmethod
    def isFileNewer(src, dst):
        if not (os.path.isfile(src) and os.path.isfile(dst)):
            return False

        # This comparison won't work with PyInstaller single-file bundles because the bundle is extracted to a temp dir
        # and every timestamp for every file in the bundle is reset to the bundle extraction/creation time.
        if getattr(sys, 'frozen', False):
            prefix = os.path.sep + "_MEI"
            if prefix in src:
                return False

        src_mtime = os.path.getmtime(os.path.abspath(src))
        dst_mtime = os.path.getmtime(os.path.abspath(dst))
        return src_mtime > dst_mtime

    @staticmethod
    def getFileSize(file_path):
        return os.path.getsize(file_path)

    @staticmethod
    def guessContentType(file_path):
        return mu.guess_content_type(file_path)

    @staticmethod
    def getFileHashes(file_path, hashes=frozenset(['md5'])):
        return hu.compute_file_hashes(file_path, hashes)

    @staticmethod
    def getCatalogTable(asset_mapping, metadata_dict=None):
        schema_name, table_name = asset_mapping.get('target_table', [None, None])
        if not (schema_name and table_name):
            metadata_dict_lower = {k.lower(): v for k, v in metadata_dict.items()}
            schema_name = metadata_dict_lower.get("schema")
            table_name = metadata_dict_lower.get("table")
        if not (schema_name and table_name):
            raise ValueError("Unable to determine target catalog table for asset type.")
        return '%s:%s' % (urlquote(schema_name), urlquote(table_name))

    @staticmethod
    def interpolateDict(src, dst, allowNone=False):
        if not (isinstance(src, dict) and isinstance(dst, dict)):
            raise ValueError("Invalid input parameter type(s): (src = %s, dst = %s), expected (dict, dict)" % (
                type(src).__name__, type(dst).__name__))

        dst = dst.copy()
        # prune None values from the src, we don't want those to be replaced with the string 'None' in the dest
        empty = [k for k, v in src.items() if v is None]
        for k in empty:
            del src[k]
        # perform the string replacement for the values in the destination dict
        for k, v in dst.items():
            try:
                value = v.format(**src)
            except KeyError:
                value = v
                if value:
                    if value.startswith('{') and value.endswith('}'):
                        value = None
            dst.update({k: value})
        # remove all None valued entries in the dest, if disallowed
        if not allowNone:
            empty = [k for k, v in dst.items() if v is None]
            for k in empty:
                del dst[k]

        return dst

    @staticmethod
    def pruneDict(src, dst, stringify=True):
        dst = dst.copy()
        for k in dst.keys():
            value = src.get(k)
            dst[k] = str(value) if (stringify and value is not None) else value
        return dst

    def getCurrentConfigFilePath(self):
        return self.override_config_file if self.override_config_file else self.getDeployedConfigFilePath()

    def getDefaultConfigFilePath(self):
        return os.path.normpath(resource_path(os.path.join("conf", self.DefaultConfigFileName)))

    def getDeployedConfigFilePath(self):
        return os.path.join(
            self.getDeployedConfigPath(), self.server.get('host', ''), self.DefaultConfigFileName)

    def getTransferStateFileName(self):
        return self.DefaultTransferStateFileName % \
               (self.DefaultTransferStateBaseName, self.server.get('host', 'localhost'))

    def getRemoteConfig(self):
        catalog_config = self.catalog.getCatalogModel()
        return catalog_config.bulk_upload

    def getUpdatedConfig(self):
        # if we are using an overridden config file, skip the update check
        if self.override_config_file:
            return

        logger.info("Checking for updated configuration...")
        remote_config = self.getRemoteConfig()
        if not remote_config:
            logger.info("Remote configuration not present, using default local configuration file.")
            return

        deployed_config_file_path = self.getDeployedConfigFilePath()
        if os.path.isfile(deployed_config_file_path):
            current_md5 = hu.compute_file_hashes(deployed_config_file_path, hashes=['md5'])['md5'][0]
        else:
            logger.info("Local config not found.")
            current_md5 = None
        tempdir = tempfile.mkdtemp(prefix="deriva_upload_")
        if os.path.exists(tempdir):
            updated_config_path = os.path.abspath(os.path.join(tempdir, DerivaUpload.DefaultConfigFileName))
            with io.open(updated_config_path, 'w', newline='\n', encoding='utf-8') as config:
                remote_config_data = json.dumps(
                    remote_config, ensure_ascii=False, sort_keys=True, separators=(',', ': '), indent=2)
                config.write(remote_config_data)
            new_md5 = hu.compute_file_hashes(updated_config_path, hashes=['md5'])['md5'][0]
            if current_md5 != new_md5:
                logger.info("Updated configuration found.")
                config = read_config(updated_config_path)
                self._update_internal_config(config)
            else:
                logger.info("Configuration is up-to-date.")
                config = None
            shutil.rmtree(tempdir, ignore_errors=True)

            return config

    def getFileStatusAsArray(self):
        result = list()
        for key in self.file_status.keys():
            item = {"File": key}
            item.update(self.file_status[key])
            result.append(item)
        return result

    @staticmethod
    def archive_preprocessing_enabled(asset_mapping):
        if not asset_mapping:
            return False
        if asset_mapping.get("archive_preprocessing_enabled", False):
            return True
        pre_processors = asset_mapping.get("pre_processors", [])
        for processor_config in pre_processors:
            processor_name = processor_config[PROCESSOR_NAME_KEY]
            processor_type = processor_config.get(PROCESSOR_TYPE_KEY)
            processor_impl = find_processor(processor_name, processor_type, bypass_whitelist=True)
            if issubclass(processor_impl, ArchiveProcessor):
                asset_mapping["archive_preprocessing_enabled"] = True
                return True
        return False

    def validateFile(self, root, path, name):
        if self.config.get("relative_path_validation", False):
            file_path = os.path.normpath(os.path.join(os.path.relpath(path, root), name))
        else:
            file_path = os.path.normpath(os.path.join(path, name))
        asset_group, asset_mapping, groupdict = self.getAssetMapping(file_path)
        if not asset_mapping:
            return None

        if self.archive_preprocessing_enabled(asset_mapping):
            final_path = os.path.abspath(os.path.normpath(groupdict.get("archive_path", path)))
        else:
            final_path = os.path.abspath(os.path.normpath(os.path.join(path, name)))

        return UploadEntry(asset_group, asset_mapping, groupdict, final_path)

    def scanDirectory(self, root, abort_on_invalid_input=False, purge_state=False):
        """

        :param root:
        :param abort_on_invalid_input:
        :param purge_state:
        :return:
        """
        root = os.path.abspath(root)
        if not os.path.isdir(root):
            raise FileNotFoundError("Invalid directory specified: [%s]" % root)
        self.loadTransferState(root, purge=purge_state)

        logger.info("Scanning files in directory [%s]..." % root)
        file_list = OrderedDict()
        for path, dirs, files in walk(root):
            for file_name in files:
                if file_name.startswith(self.DefaultTransferStateBaseName):
                    continue
                file_path = os.path.normpath(os.path.join(path, file_name))
                upload_entry = self.validateFile(root, path, file_name)
                if not upload_entry:
                    logger.info("Skipping file: [%s] -- Invalid file type or directory location." % file_path)
                    self.skipped_files.add(file_path)
                    if abort_on_invalid_input:
                        raise DerivaUploadError("Invalid input detected, aborting.")
                else:
                    asset_group = upload_entry.asset_group
                    group_list = file_list.get(asset_group, {})
                    group_list.update({upload_entry.path: upload_entry})
                    file_list[asset_group] = group_list

        # make sure that file entries in both self.file_list and self.file_status are ordered by the declared order of
        # the asset_mapping for the file
        for group in sorted(file_list.keys()):
            self.file_list[group] = file_list[group]
            for upload_entry in file_list[group].values():
                file_path = upload_entry.path
                logger.info("Including %s: [%s]." %
                            ("directory (for archive)" if self.archive_preprocessing_enabled(
                                upload_entry.asset_mapping) else "file", file_path))
                status = self.getTransferStateStatus(file_path)
                if status:
                    self.file_status[file_path] = FileUploadState(UploadState.Paused, status).asdict()
                else:
                    self.file_status[file_path] = FileUploadState(UploadState.Pending, "Pending").asdict()

    def getAssetMapping(self, file_path):
        """
        :param file_path:
        :return:
        """
        asset_group = -1
        for asset_type in self.asset_mappings:
            asset_group += 1
            groupdict = dict()
            dir_pattern = asset_type.get('dir_pattern', '')
            ext_pattern = asset_type.get('ext_pattern', '')
            file_pattern = asset_type.get('file_pattern', '')
            path = file_path.replace("\\", "/")
            if dir_pattern:
                match = re.search(dir_pattern, path)
                if not match:
                    logger.debug("The dir_pattern \"%s\" failed to match the input path [%s]" % (dir_pattern, path))
                    continue
                groupdict.update(match.groupdict())
            if ext_pattern:
                if self.archive_preprocessing_enabled(asset_type):
                    logger.warning("The 'ext_pattern' parameter is not compatible when archive preprocessing "
                                   "is enabled. Only input directories matching 'dir_pattern' are supported.")
                    continue
                match = re.search(ext_pattern, path, re.IGNORECASE)
                if not match:
                    logger.debug("The ext_pattern \"%s\" failed to match the input path [%s]" % (ext_pattern, path))
                    continue
                groupdict.update(match.groupdict())
            if file_pattern:
                if self.archive_preprocessing_enabled(asset_type):
                    logger.warning("The 'file_pattern' parameter is not compatible when archive preprocessing "
                                   "is enabled. Only input directories matching 'dir_pattern' are supported.")
                    continue
                match = re.search(file_pattern, path)
                if not match:
                    logger.debug("The file_pattern \"%s\" failed to match the input path [%s]" % (file_pattern, path))
                    continue
                groupdict.update(match.groupdict())

            return asset_group, asset_type, groupdict

        return None, None, None

    def uploadFiles(self, status_callback=None, file_callback=None):
        completed = 0
        for group, assets in self.file_list.items():
            if self.cancelled:
                break
            for entry in assets.values():
                if self.cancelled:
                    self.file_status[entry.path] = FileUploadState(UploadState.Cancelled, "Cancelled by user").asdict()
                    break
                try:
                    self.file_status[entry.path] = FileUploadState(UploadState.Running, "In-progress").asdict()
                    if status_callback:
                        status_callback()
                    result = self.uploadFile(entry.path,
                                             entry.asset_mapping,
                                             entry.groupdict,
                                             file_callback or self.defaultFileCallback)
                    if self.cancelled:
                        self.file_status[entry.path] = FileUploadState(UploadState.Cancelled,
                                                                       "Cancelled by user").asdict()
                        break
                    else:
                        self.file_status[entry.path] = FileUploadState(UploadState.Success, "Complete", result).asdict()
                        completed += 1
                except HatracJobPaused:
                    status = self.getTransferStateStatus(entry.path)
                    if status:
                        self.file_status[entry.path] = FileUploadState(
                            UploadState.Paused, "Paused: %s" % status).asdict()
                    continue
                except HatracJobTimeout:
                    status = self.getTransferStateStatus(entry.path)
                    if status:
                        self.file_status[entry.path] = FileUploadState(UploadState.Timeout, "Timeout").asdict()
                    continue
                except HatracJobAborted:
                    self.file_status[entry.path] = FileUploadState(UploadState.Aborted, "Aborted by user").asdict()
                except:
                    logger.debug("Unexpected exception", exc_info=sys.exc_info())
                    (etype, value, traceback) = sys.exc_info()
                    self.file_status[entry.path] = FileUploadState(UploadState.Failed, format_exception(value)).asdict()
                self.delTransferState(entry.path)
                if status_callback:
                    status_callback()

        failed_uploads = dict()
        try:
            for key, value in self.file_status.items():
                if (value["State"] == UploadState.Failed) or (value["State"] == UploadState.Timeout):
                    failed_uploads[key] = value["Status"]

            if self.skipped_files:
                logger.warning("The following %d file(s) were skipped because they did not satisfy the matching "
                               "criteria of the configuration:\n\n%s\n" %
                               (len(self.skipped_files), '\n'.join(sorted(self.skipped_files))))

            if failed_uploads:
                logger.warning("The following %d file(s) failed to upload due to errors:\n\n%s\n" %
                               (len(failed_uploads), '\n'.join(["%s -- %s" % (key, failed_uploads[key])
                                                                for key in sorted(failed_uploads.keys())])))
                raise RuntimeError("%s file(s) failed to upload due to errors." % len(failed_uploads))
        finally:
            logger.info("File upload processing completed: %s files were uploaded successfully, "
                        "%s files failed to upload due to errors, "
                        "%s files were skipped because they did not satisfy the matching criteria of the configuration."
                        % (completed, len(failed_uploads), len(self.skipped_files)))

        return self.file_status

    def uploadFile(self, file_path, asset_mapping, match_groupdict, callback=None):
        """
        Primary API subclass function.
        :param file_path:
        :param asset_mapping:
        :param match_groupdict:
        :param callback:
        :return:
        """
        logger.info("Processing: [%s]" % file_path)

        if asset_mapping.get("asset_type", "file") == "table":
            return self._uploadTable(file_path, asset_mapping, match_groupdict)
        else:
            return self._uploadAsset(file_path, asset_mapping, match_groupdict, callback)

    def _uploadAsset(self, file_path, asset_mapping, match_groupdict, callback=None):

        # 1. Populate initial file metadata from directory scan pattern matches
        self._initFileMetadata(file_path, asset_mapping, match_groupdict)

        # 2. Execute any configured preprocessors
        self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=PRE_PROCESSORS_KEY)
        if PROCESSOR_MODIFIED_FILE_PATH_KEY in self.processor_output:
            file_path = self.processor_output[PROCESSOR_MODIFIED_FILE_PATH_KEY]
        self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
        logger.info("Computed metadata for: [%s]." % file_path)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Current metadata: %s" % self.metadata)

        # 3. Compute checksum(s) for current file and add to metadata
        logger.info("Computing checksums for file: [%s]. Please wait..." % file_path)
        hashes = self.getFileHashes(file_path, asset_mapping.get('checksum_types', ['md5', 'sha256']))
        for alg, checksum in hashes.items():
            alg = alg.lower()
            self.metadata[alg] = checksum[0]
            self.metadata[alg + "_base64"] = checksum[1]
        if self.cancelled:
            return

        # 4. Populate additional metadata by querying the catalog
        self._queryFileMetadata(asset_mapping)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Updated metadata: %s" % self.metadata)

        # 5. If "create_record_before_upload" specified in asset_mapping, check for an existing record, creating a new
        #    one if necessary. Otherwise, delay this logic until after the file upload.
        result = record = None
        if stob(asset_mapping.get("create_record_before_upload", False)):
            record = self._getFileRecord(asset_mapping)

        # 6. Perform the Hatrac upload
        self._getFileHatracMetadata(asset_mapping)
        hatrac_options = asset_mapping.get("hatrac_options", {})
        file_size = self.metadata["file_size"]
        versioned_uri = \
            self._hatracUpload(self.metadata["URI"],
                               file_path,
                               md5=self.metadata.get("md5_base64"),
                               sha256=self.metadata.get("sha256_base64"),
                               content_type=self.guessContentType(file_path),
                               content_disposition=self.metadata.get("content-disposition"),
                               chunked=True if (file_size > DEFAULT_CHUNK_SIZE or file_size == 0) else False,
                               create_parents=stob(hatrac_options.get("create_parents", True)),
                               allow_versioning=stob(hatrac_options.get("allow_versioning", True)),
                               callback=callback)
        logger.debug("Hatrac upload successful. Result object URI: %s" % versioned_uri)
        versioned_uris = True
        if "versioned_uris" in hatrac_options:
            versioned_uris = stob(hatrac_options.get("versioned_uris", True))
        if "versioned_urls" in hatrac_options:
            versioned_uris = stob(hatrac_options.get("versioned_urls", True))
        if versioned_uris:
            self.metadata["URI"] = versioned_uri
        else:
            self.metadata["URI"] = versioned_uri.rsplit(":")[0]
        self.metadata["URI_urlencoded"] = urlquote(self.metadata["URI"])

        # 7. Check for an existing record and create a new one if necessary
        if not record:
            record, result = self._getFileRecord(asset_mapping)

        # 8. Update an existing record, if necessary
        column_map = asset_mapping.get("column_map", {})
        updated_record = self.interpolateDict(self.metadata, column_map)
        if updated_record != record:
            record_update_template = asset_mapping.get("record_update_template")
            require_record_update_template = stob(asset_mapping.get("require_record_update_template", False))
            if require_record_update_template and not record_update_template:
                raise DerivaUploadCatalogUpdateError(
                    "A required 'record_update_template' parameter for this asset mapping could not be found in the "
                    "configuration. The record will not be updated.")
            logger.info("Updating catalog for file [%s]" % self.getFileDisplayName(file_path))
            result = self._catalogRecordUpdate(self.metadata['target_table'],
                                               record,
                                               updated_record,
                                               record_update_template)[0]
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Updated catalog for file [%s]: %s" % (self.getFileDisplayName(file_path), result))
            record, result = self._getFileRecord(asset_mapping)

        # 9. Execute any configured post_processors
        self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=POST_PROCESSORS_KEY)

        return result

    def _uploadTable(self, file_path, asset_mapping, match_groupdict, callback=None):
        if self.cancelled:
            return None

        self._initFileMetadata(file_path, asset_mapping, match_groupdict)
        self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=PRE_PROCESSORS_KEY)
        try:
            default_columns = asset_mapping.get("default_columns")
            if not default_columns:
                default_columns = self.catalog.getDefaultColumns({}, self.metadata['target_table'])
            default_param = ('?defaults=%s' % ','.join(default_columns)) if len(default_columns) > 0 else ''
            file_ext = self.metadata['file_ext']
            file_ext = file_ext.lower()
            if file_ext == 'csv':
                headers = {'content-type': 'text/csv'}
            elif file_ext == 'json':
                headers = {'content-type': 'application/json'}
            else:
                raise DerivaUploadCatalogCreateError("Unsupported file type for catalog bulk upload: %s" % file_ext)
            with open(file_path, "rb") as fp:
                result = self.catalog.post(
                    '/entity/%s%s' % (self.metadata['target_table'], default_param), fp, headers=headers)
                return result
        except:
            (etype, value, traceback) = sys.exc_info()
            raise DerivaUploadCatalogCreateError(format_exception(value))
        finally:
            self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=POST_PROCESSORS_KEY)

    def _getFileRecord(self, asset_mapping):
        """
        Helper function that queries the catalog to get a record linked to the asset, or create it if it doesn't exist.
        :return: the file record
        """
        record = None
        column_map = asset_mapping.get("column_map", {})
        rqt = asset_mapping['record_query_template']
        try:
            path = rqt.format(**self.metadata)
        except KeyError as e:
            raise DerivaUploadConfigurationError("Record query template substitution error: %s" % format_exception(e))
        result = self.catalog.get(path).json()
        if result:
            record = result[0]
            self._updateFileMetadata(record, no_overwrite=True)
            return self.pruneDict(record, column_map), record
        else:
            row = self.interpolateDict(self.metadata, column_map)
            result = self._catalogRecordCreate(self.metadata['target_table'], row)
            if result:
                record = result[0]
                self._updateFileMetadata(record)
            return self.interpolateDict(self.metadata, column_map, allowNone=True), record

    def _urlEncodeMetadata(self, safe_overrides=None):
        urlencoded = dict()
        if not safe_overrides:
            safe_overrides = dict()
        for k, v in self.metadata.items():
            if k.endswith("_urlencoded"):
                continue
            urlencoded[k + "_urlencoded"] = urlquote(str(v), safe_overrides.get(k, ""))
        self._updateFileMetadata(urlencoded)

    def _initFileMetadata(self, file_path, asset_mapping, match_groupdict):
        self.metadata.clear()
        self._updateFileMetadata(match_groupdict)
        self.metadata['target_table'] = self.getCatalogTable(asset_mapping, match_groupdict)

        self.metadata["file_name"] = self.getFileDisplayName(file_path)
        self.metadata["file_size"] = self.getFileSize(file_path)
        if "file_ext" not in self.metadata:
            self.metadata["file_ext"] = "".join(pathlib.PurePath(file_path).suffixes)
        self.metadata["base_path"] = os.path.dirname(file_path)
        self.metadata["base_name"] = self.metadata["file_name"].rsplit(
            self.metadata["file_ext"])[0] if self.metadata["file_ext"] else self.metadata["file_name"]

        time = datetime.datetime.now()
        self.metadata["_upload_year_"] = time.year
        self.metadata["_upload_month_"] = time.month
        self.metadata["_upload_day_"] = time.day
        self.metadata["_upload_time_"] = time.timestamp()
        self.metadata["_identity_id"] = self.identity.get("id", "anonymous")
        self.metadata["_identity_display_name"] = self.identity.get("display_name")
        self.metadata["_identity_full_name"] = self.identity.get("full_name")
        self.metadata["_identity_email"] = self.identity.get("email")

        self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))

    def _updateFileMetadata(self, src, strict=False, no_overwrite=False):
        if not (isinstance(src, dict)):
            raise ValueError("Invalid input parameter type(s): (src = %s), expected (dict)" % type(src).__name__)
        dst = src.copy()
        for k in src.keys():
            if strict:
                if k in UploadMetadataReservedKeyNames:
                    logger.warning("Context metadata update specified reserved key name [%s], "
                                   "ignoring value: %s " % (k, src[k]))
                    del dst[k]
                    continue
            # don't overwrite any existing metadata field
            if no_overwrite:
                if k in self.metadata:
                    del dst[k]
        self.metadata.update(dst)

    def _queryFileMetadata(self, asset_mapping):
        """
        Helper function that queries the catalog to get required metadata for a given file/asset
        """
        metadata_queries = asset_mapping.get("metadata_query_templates", [])
        if logger.isEnabledFor(logging.DEBUG) and metadata_queries:
            logger.debug("Querying catalog for additional metadata...")
        for uri in metadata_queries:
            try:
                path = uri.format(**self.metadata)
            except KeyError as e:
                raise RuntimeError("Metadata query template substitution error: %s" % format_exception(e))
            result = self.catalog.get(path).json()
            if result:
                self._updateFileMetadata(result[0], True)
            else:
                raise RuntimeError("Metadata query did not return any results: %s" % path)

        self._getFileExtensionMetadata(self.metadata.get("file_ext"))

        for k, v in asset_mapping.get("column_value_templates", {}).items():
            try:
                self.metadata[k] = v.format(**self.metadata)
            except KeyError as e:
                logger.warning("Column value template substitution error: %s" % format_exception(e))
                continue

        self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))

    def _getFileExtensionMetadata(self, ext):
        ext_map = self.config.get("file_ext_mappings", {})
        entry = ext_map.get(ext)
        if entry:
            self._updateFileMetadata(entry)

    def _getFileHatracMetadata(self, asset_mapping):
        try:
            hatrac_templates = asset_mapping["hatrac_templates"]
            # URI is required
            self.metadata["URI"] = hatrac_templates["hatrac_uri"].format(**self.metadata)
            # overridden content-disposition is optional
            content_disposition = hatrac_templates.get("content-disposition")
            if content_disposition:
                filename = content_disposition.format(**self.metadata)
            else:
                filename = urlparse(self.metadata["URI"]).path.rsplit("/", 1)[-1]

            sanitized_filename, sanitized_content_disp = \
                self._validateHatracFilename(filename, asset_mapping.get("hatrac_options", {}))
            if content_disposition:
                self.metadata["content-disposition"] = sanitized_content_disp
            else:
                self.metadata["URI"] = self.metadata["URI"].replace(filename, sanitized_filename)

            self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
        except KeyError as e:
            raise DerivaUploadConfigurationError("Hatrac template substitution error: %s" % format_exception(e))

    def _validateHatracFilename(self, filename, hatrac_options):
        if not filename:
            return None

        sanitize = hatrac_options.get("sanitize_filenames", True)
        pattern = hatrac_options.get("sanitize_filenames_pattern")
        is_content_disp = re.match(r"filename\*?=['\"]?(?:UTF-\d['\"]*)?([^;\r\n\"']*)['\"]?;?", filename)
        if is_content_disp:
            filename = is_content_disp.group(1)
        pattern = pattern if pattern else "[^a-zA-Z0-9_.-]"
        sanitized_filename = urlquote(re.sub(pattern, "_", filename)) if sanitize else filename
        if is_content_disp:
            content_disp = is_content_disp.string.replace(filename, sanitized_filename)
        else:
            content_disp = "filename*=UTF-8''" + sanitized_filename

        return sanitized_filename, content_disp

    def _hatracUpload(self,
                      uri,
                      file_path,
                      md5=None,
                      sha256=None,
                      content_type=None,
                      content_disposition=None,
                      chunked=True,
                      create_parents=True,
                      allow_versioning=True,
                      callback=None):

        # check if there is already an in-progress transfer for this file,
        # and if so, that the local file has not been modified since the original upload job was created
        can_resume = False
        transfer_state = self.getTransferState(file_path)
        if transfer_state:
            content_md5 = transfer_state.get("content-md5")
            content_sha256 = transfer_state.get("content-sha256")
            if content_md5 or content_sha256:
                if (md5 == content_md5) or (sha256 == content_sha256):
                    can_resume = True

        if transfer_state and can_resume:
            logger.info("Resuming upload (%s) of file: [%s] to host %s. Please wait..." % (
                self.getTransferStateStatus(file_path), file_path, transfer_state.get("host")))
            path = transfer_state["target"]
            job_id = transfer_state['url'].rsplit("/", 1)[1]
            if not (transfer_state["total"] == transfer_state["completed"]):
                self.store.put_obj_chunked(path,
                                           file_path,
                                           job_id,
                                           callback=callback,
                                           start_chunk=transfer_state["completed"],
                                           cancel_job_on_error=False)
            return self.store.finalize_upload_job(path, job_id)
        else:
            logger.info("Uploading file: [%s] to host %s. Please wait..." % (file_path, self.server_url))
            return self.store.put_loc(uri,
                                      file_path,
                                      md5=md5,
                                      sha256=sha256,
                                      content_type=content_type,
                                      content_disposition=content_disposition,
                                      chunked=chunked,
                                      create_parents=create_parents,
                                      allow_versioning=allow_versioning,
                                      callback=callback,
                                      cancel_job_on_error=False)

    def _get_catalog_table_columns(self, table):
        table_columns = set()
        catalog_table_metadata = self.catalog_metadata["table_metadata"]
        table_metadata = catalog_table_metadata.get(table)
        if table_metadata:
            table_columns = table_metadata.get("table_columns")
        if not table_columns:
            table_columns = self.catalog.getTableColumns(table)
            catalog_table_metadata.update({table: {"table_columns": table_columns}})
        return table_columns

    def _validate_catalog_row_columns(self, row, table):
        return set(row.keys()) - self._get_catalog_table_columns(table)

    def _validate_row_key_constraints(self, catalog_table, row):
        logger.debug("Validating row key constraints for %s: %s" % (catalog_table, row))
        if not self.catalog_model:
            logger.debug("Fetching catalog model...")
            self.catalog_model = self.catalog.getCatalogModel()
        schema_name, table_name = self.catalog.splitQualifiedCatalogName(catalog_table)
        schema = self.catalog_model.schemas.get(schema_name)
        table = schema.tables.get(table_name)
        non_null_correlations = {cname for cname, cval in row.items() if cval is not None}
        for key in table.keys:
            if set(key.unique_columns.elements).issubset(non_null_correlations):
                logger.debug("%s is a subset of non-null correlations %s" %
                             (set(key.unique_columns.elements), non_null_correlations))
                return True  # it is safe
            else:
                logger.debug("%s is not a subset of non-null correlations %s" %
                             (set(key.unique_columns.elements), non_null_correlations))
        return False  # it is not safe

    def _get_catalog_default_columns(self, row, table, exclude=None, quote_url=True):
        columns = self._get_catalog_table_columns(table)
        if isinstance(exclude, list):
            for col in exclude:
                columns.remove(col)

        defaults = []
        supplied_columns = row.keys()
        for col in columns:
            if col not in supplied_columns:
                defaults.append(urlquote(col, safe='') if quote_url else col)

        return defaults

    def _catalogRecordCreate(self, catalog_table, row, default_columns=None):
        """

        :param catalog_table:
        :param row:
        :param default_columns:
        :return:
        """
        if self.cancelled:
            return None

        try:
            missing = self._validate_catalog_row_columns(row, catalog_table)
            if missing:
                raise ValueError(
                    "Unable to update catalog entry because one or more specified columns do not exist in the "
                    "target table: [%s]" % ','.join(missing))
            if not default_columns:
                default_columns = self._get_catalog_default_columns(row, catalog_table)
            default_param = ('?defaults=%s' % ','.join(default_columns)) if len(default_columns) > 0 else ''
            # for default in default_columns:
            #    row[default] = None
            create_uri = '/entity/%s%s' % (catalog_table, default_param)
            logger.debug(
                "Attempting catalog record create [%s] with data: %s" % (create_uri, json.dumps(row)))
            return self.catalog.post(create_uri, json=[row]).json()
        except:
            (etype, value, traceback) = sys.exc_info()
            raise DerivaUploadCatalogCreateError(format_exception(value))

    def _catalogRecordUpdate(self, catalog_table, old_row, new_row, record_update_template=None):
        """

        :param catalog_table:
        :param new_row:
        :param old_row:
        :return:
        """
        if self.cancelled:
            return None

        try:
            keys = sorted(list(new_row.keys()))
            old_keys = sorted(list(old_row.keys()))
            if keys != old_keys:
                raise RuntimeError("Cannot update catalog - "
                                   "new row column list and old row column list do not match: New: %s != Old: %s" %
                                   (keys, old_keys))
            o_keys = ','.join(["o%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))])
            n_keys = ','.join(["n%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))])
            update_row = {
                'o%d' % i: old_row[keys[i]]
                for i in range(len(keys))
            }
            update_row.update({
                'n%d' % i: new_row[keys[i]]
                for i in range(len(keys))
            })
            if record_update_template:
                update_uri = record_update_template.format(**self.metadata)
                update_row = new_row
            else:
                update_uri = '/attributegroup/%s/%s;%s' % (catalog_table, o_keys, n_keys)
                if self.config.get("strict_update_check", True) and not \
                        self._validate_row_key_constraints(catalog_table, old_row):
                    raise ValueError(
                        "Potential unsafe attributegroup update [%s]: at least one pre-existing, non-null correlation "
                        "key is required. Old values: %s, New values: %s" %
                        (update_uri, json.dumps(old_row), json.dumps(new_row)))
            logger.debug(
                "Attempting catalog record update [%s] with data: %s" % (update_uri, json.dumps(update_row)))
            return self.catalog.put(update_uri, json=[update_row]).json()
        except:
            (etype, value, traceback) = sys.exc_info()
            raise DerivaUploadCatalogUpdateError(format_exception(value))

    def _execute_processors(self, file_path, asset_mapping, match_groupdict,
                            processor_list=PRE_PROCESSORS_KEY, **kwargs):
        processors = asset_mapping.get(processor_list, [])
        if processors:
            for processor_config in processors:
                processor_name = processor_config[PROCESSOR_NAME_KEY]
                processor_type = processor_config.get(PROCESSOR_TYPE_KEY)
                processor_params = processor_config.get(PROCESSOR_PARAMS_KEY)
                try:
                    processor_impl = find_processor(processor_name, processor_type, bypass_whitelist=True)
                    processor = processor_impl(
                        processor_params=processor_params,
                        file_path=file_path,
                        asset_mapping=asset_mapping,
                        match_groupdict=match_groupdict,
                        metadata=self.metadata,
                        processor_output=self.processor_output,
                        **kwargs)
                    proc_class = processor.__class__.__module__
                    proc_name = processor.__class__.__name__
                    if processor_params is not None and processor_params.get(
                            PROCESSOR_REQUIRES_METADATA_QUERY_KEY, False):
                        self._queryFileMetadata(asset_mapping)
                    logger.debug("Attempting to execute upload processor class %s from module: %s" %
                                 (proc_name, proc_class))
                    output = processor.process()
                    if isinstance(output, dict):
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug("%s output context: %s" % (proc_name, output))
                except:
                    (etype, value, traceback) = sys.exc_info()
                    raise DerivaUploadError(format_exception(value))

    def defaultFileCallback(self, **kwargs):
        completed = kwargs.get("completed")
        total = kwargs.get("total")
        file_path = kwargs.get("file_path")
        file_name = os.path.basename(file_path) if file_path else ""
        job_info = kwargs.get("job_info", {})
        job_info.update()
        if completed and total:
            file_name = " [%s]" % file_name
            job_info.update({"completed": completed, "total": total, "host": kwargs.get("host")})
            status = "Uploading file%s: %d%% complete" % (
                file_name, round(((float(completed) / float(total)) % 100) * 100))
            self.setTransferState(file_path, job_info)
        else:
            summary = kwargs.get("summary", "")
            file_name = "Uploaded file: [%s] " % file_name
            status = file_name  # + summary
        if status:
            # logger.debug(status)
            pass
        if self.cancelled:
            return -1

        return True

    @staticmethod
    def find_file_in_dir_hierarchy(filename, path):
        """ Find all instances of a filename in the entire directory hierarchy specified by path.
        """
        file_paths = set()
        # First, descend from the base path looking for filename in all sub dirs
        for root, dirs, files in walk(path, followlinks=True):
            if filename in files:
                # found a file
                found_path = os.path.normcase(os.path.join(root, filename))
                file_paths.add(os.path.normcase(os.path.realpath(found_path)))
                continue

        # Next, ascend from the base path looking for the same filename in all parent dirs
        current = path
        while True:
            parent = os.path.dirname(current)
            if parent == current:
                break
            for entry in scandir(parent):
                if (entry.name == filename) and entry.is_file():
                    file_paths.add(os.path.normcase(os.path.realpath(os.path.normcase(entry.path))))
            current = parent

        return file_paths

    def delete_dependent_locks(self, directory):
        for path in self.find_file_in_dir_hierarchy(self.getTransferStateFileName(), directory):
            logger.info("Attempting to delete an existing transfer state file (dependent lock) at: [%s]" % path)
            try:
                os.remove(path)
            except OSError as e:
                logger.warning("Unable to delete transfer state file [%s]: %s" % (path, format_exception(e)))

    def acquire_dependent_locks(self, directory):
        for path in self.find_file_in_dir_hierarchy(self.getTransferStateFileName(), directory):
            logger.info("Attempting to acquire a dependent lock in [%s]" % os.path.dirname(path))
            try:
                transfer_state_lock = lock_file(path, 'r+')
                transfer_state_fh = transfer_state_lock.acquire(timeout=0, fail_when_locked=True)
                self.transfer_state_locks.update({path: {"lock": transfer_state_lock, "handle": transfer_state_fh}})
            except Exception as e:
                raise DerivaUploadError("Unable to acquire resource lock for directory [%s]. "
                                        "Multiple upload processes cannot operate within the same directory hierarchy. "
                                        "%s" % (os.path.dirname(path), format_exception(e)))

    def loadTransferState(self, directory, purge=False):
        transfer_state_file_path = os.path.normcase(os.path.join(directory, self.getTransferStateFileName()))
        if purge:
            self.delete_dependent_locks(directory)
        self.acquire_dependent_locks(directory)
        try:
            if not os.path.isfile(transfer_state_file_path):
                with lock_file(transfer_state_file_path, "w") as tsfp:
                    json.dump(self.transfer_state, tsfp)

            transfer_state_lock = self.transfer_state_locks.get(transfer_state_file_path)
            if transfer_state_lock:
                self.transfer_state_fh = transfer_state_lock["handle"]
            else:
                transfer_state_lock = lock_file(transfer_state_file_path, 'r+')
                self.transfer_state_fh = transfer_state_lock.acquire(timeout=0, fail_when_locked=True)
                self.transfer_state_locks.update(
                    {directory: {"lock": transfer_state_lock, "handle": self.transfer_state_fh}})
            self.transfer_state = json.load(self.transfer_state_fh, object_pairs_hook=OrderedDict)
        except Exception as e:
            raise DerivaUploadError("Unable to acquire resource lock for directory [%s]. "
                                    "Multiple upload processes cannot operate within the same directory hierarchy. %s"
                                    % (directory, format_exception(e)))

    def getTransferState(self, file_path):
        return self.transfer_state.get(file_path)

    def setTransferState(self, file_path, transfer_state):
        self.transfer_state[file_path] = transfer_state
        self.writeTransferState()

    def delTransferState(self, file_path):
        transfer_state = self.getTransferState(file_path)
        if transfer_state:
            del self.transfer_state[file_path]
        self.writeTransferState()

    def writeTransferState(self):
        if not self.transfer_state_fh:
            return
        try:
            self.transfer_state_fh.seek(0, 0)
            self.transfer_state_fh.truncate()
            json.dump(self.transfer_state, self.transfer_state_fh, indent=2)
            self.transfer_state_fh.flush()
            os.fsync(self.transfer_state_fh.fileno())
        except Exception as e:
            logger.warning("Unable to write transfer state file: %s" % format_exception(e))

    def cleanupTransferState(self):
        if self.transfer_state_fh and not self.transfer_state_fh.closed:
            try:
                self.transfer_state_fh.flush()
                os.fsync(self.transfer_state_fh.fileno())
            except Exception as e:
                logger.warning("Unable to flush/close transfer state file: %s" % format_exception(e))
            finally:
                for entry in self.transfer_state_locks.values():
                    lock = entry.get("lock")
                    if lock and not lock.fh.closed:
                        lock.release()
                self.transfer_state_locks.clear()
                self.transfer_state_fh = None

    def getTransferStateStatus(self, file_path):
        transfer_state = self.getTransferState(file_path)
        if transfer_state:
            return "%d%% complete" % (
                round(((float(transfer_state["completed"]) / float(transfer_state["total"])) % 100) * 100))
        return None


class GenericUploader(DerivaUpload):

    def __init__(self, config_file=None, credential_file=None, server=None, dcctx_cid=None):
        DerivaUpload.__init__(self,
                              config_file=config_file,
                              credential_file=credential_file,
                              server=server,
                              dcctx_cid=dcctx_cid)

    @classmethod
    def getVersion(cls):
        return VERSION

    @classmethod
    def getConfigPath(cls):
        return "~/.deriva/upload/"

    @classmethod
    def getServers(cls):
        return read_config(os.path.join(
            cls.getDeployedConfigPath(), cls.DefaultServerListFileName), create_default=True, default=[])

    @classmethod
    def setServers(cls, servers):
        return write_config(os.path.join(cls.getDeployedConfigPath(), cls.DefaultServerListFileName), servers)



class DerivaUploadError (RuntimeError):
    pass


class DerivaUploadConfigurationError (ValueError):
    pass


class DerivaUploadCatalogCreateError (RuntimeError):
    pass


class DerivaUploadCatalogUpdateError (RuntimeError):
    pass



import os
import errno
import certifi
import requests
from deriva.core import urlsplit, get_new_requests_session, stob, make_dirs, DEFAULT_SESSION_CONFIG
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
    DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
from deriva.transfer.download.processors.base_processor import BaseProcessor, \
    LOCAL_PATH_KEY, FILE_SIZE_KEY, SOURCE_URL_KEY
from bdbag import bdbag_ro as ro


class BaseQueryProcessor(BaseProcessor):
    """
    Base class for QueryProcessor classes
    """
    HEADERS = {'Connection': 'keep-alive'}

    def __init__(self, envars=None, **kwargs):
        super(BaseQueryProcessor, self).__init__(envars, **kwargs)
        self.catalog = kwargs["catalog"]
        self.store = kwargs["store"]
        self.base_path = kwargs["base_path"]
        self.query = self.parameters["query_path"]
        if self.envars:
            self.query = self.query.format(**self.envars)
        self.sub_path = self.parameters.get("output_path")
        self.output_filename = self.parameters.get("output_filename")
        self.store_base = kwargs.get("store_base", "/hatrac/")
        self.is_bag = kwargs.get("bag", False)
        self.sessions = kwargs.get("sessions", dict())
        self.content_type = "application/octet-stream"
        self.url = ''.join([self.catalog.get_server_uri(), self.query])
        self.ro_file_provenance = stob(self.parameters.get("ro_file_provenance", False if not self.is_bag else True))
        self.ro_manifest = self.kwargs.get("ro_manifest")
        self.ro_author_name = self.kwargs.get("ro_author_name")
        self.ro_author_orcid = self.kwargs.get("ro_author_orcid")
        self.output_relpath = None
        self.output_abspath = None
        self.paged_query = self.parameters.get("paged_query", False)
        self.paged_query_size = self.parameters.get("paged_query_size", 100000)
        self.paged_query_sort_columns = self.parameters.get("paged_query_sort_columns", ["RID"])

    def process(self):
        resp = self.catalogQuery(headers={'accept': self.content_type})
        if os.path.isfile(self.output_abspath):
            if self.ro_manifest and self.ro_file_provenance:
                ro.add_file_metadata(self.ro_manifest,
                                     source_url=self.url,
                                     local_path=self.output_relpath,
                                     media_type=self.content_type,
                                     retrieved_on=ro.make_retrieved_on(),
                                     retrieved_by=ro.make_retrieved_by(self.ro_author_name,
                                                                       orcid=self.ro_author_orcid),
                                     bundled_as=ro.make_bundled_as())
            self.outputs.update({self.output_relpath: {LOCAL_PATH_KEY: self.output_abspath,
                                                       FILE_SIZE_KEY: os.path.getsize(self.output_abspath),
                                                       SOURCE_URL_KEY: self.url}})
        return self.outputs

    def catalogQuery(self, headers=None, as_file=True):
        if not headers:
            headers = self.HEADERS.copy()
        else:
            headers.update(self.HEADERS)

        if as_file:
            output_dir = os.path.dirname(self.output_abspath)
            make_dirs(output_dir)
        try:
            if as_file:
                return self.catalog.getAsFile(self.query, self.output_abspath,
                                              headers=headers,
                                              callback=self.callback,
                                              delete_if_empty=True,
                                              paged=self.paged_query,
                                              page_size=self.paged_query_size,
                                              page_sort_columns=self.paged_query_sort_columns)
            else:
                return self.catalog.get(self.query, headers=headers).json()
        except requests.HTTPError as e:
            if e.response.status_code == 401:
                raise DerivaDownloadAuthenticationError(e)
            if e.response.status_code == 403:
                raise DerivaDownloadAuthorizationError(e)
            if as_file:
                os.remove(self.output_abspath)
            raise DerivaDownloadError("Error executing catalog query: %s" % e)
        except Exception:
            if as_file:
                os.remove(self.output_abspath)
            raise

    def headForHeaders(self, url, raise_for_status=False):
        store = self.getHatracStore(url)
        if store:
            r = store.head(url, headers=self.HEADERS)
            if raise_for_status:
                r.raise_for_status()
            headers = r.headers
        else:
            url = self.getExternalUrl(url)
            session = self.getExternalSession(urlsplit(url).hostname)
            r = session.head(url, headers=self.HEADERS)
            if raise_for_status:
                r.raise_for_status()
            headers = r.headers

        return headers

    def getHatracStore(self, url):
        urlparts = urlsplit(url)
        if not urlparts.path.startswith(self.store_base):
            return None
        if url.startswith(self.store_base):
            return self.store
        else:
            serverURI = urlparts.scheme + "://" + urlparts.netloc
            if serverURI == self.store.get_server_uri():
                return self.store
            else:
                # do we need to deal with the possibility of a fully qualified URL referencing a different hatrac host?
                raise DerivaDownloadConfigurationError(
                    "Got a reference to a Hatrac server [%s] that is different from the expected Hatrac server: %s" % (
                        serverURI, self.store.get_server_uri))

    def getExternalUrl(self, url):
        urlparts = urlsplit(url)
        if urlparts.path.startswith(self.store_base):
            path_only = url.startswith(self.store_base)
            server_uri = urlparts.scheme + "://" + urlparts.netloc
            if server_uri == self.store.get_server_uri() or path_only:
                url = ''.join([self.store.get_server_uri(), url]) if path_only else url
        else:
            if not (urlparts.scheme and urlparts.netloc):
                urlparts = urlsplit(self.catalog.get_server_uri())
                server_uri = urlparts.scheme + "://" + urlparts.netloc
                url = ''.join([server_uri, url])

        return url

    def getExternalSession(self, host):
        sessions = self.sessions
        auth_params = self.kwargs.get("auth_params", dict())
        cookies = auth_params.get("cookies")
        auth_url = auth_params.get("auth_url")
        login_params = auth_params.get("login_params")
        session_config = self.kwargs.get("session_config")

        session = sessions.get(host)
        if session is not None:
            return session

        if not session_config:
            session_config = DEFAULT_SESSION_CONFIG
        session = get_new_requests_session(session_config=session_config)

        if cookies:
            session.cookies.update(cookies)
        if login_params and auth_url:
            r = session.post(auth_url, data=login_params, verify=certifi.where())
            if r.status_code > 203:
                raise DerivaDownloadError(
                    'GetExternalSession Failed with Status Code: %s\n%s\n' % (r.status_code, r.text))

        sessions[host] = session
        return session

    def create_default_paths(self):
        self.output_relpath, self.output_abspath = self.create_paths(self.base_path,
                                                                     sub_path=self.sub_path,
                                                                     filename=self.output_filename,
                                                                     ext=self.ext,
                                                                     is_bag=self.is_bag,
                                                                     envars=self.envars)

    def __del__(self):
        for session in self.sessions.values():
            session.close()


class CSVQueryProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(CSVQueryProcessor, self).__init__(envars, **kwargs)
        self.ext = ".csv"
        self.content_type = "text/csv"
        self.create_default_paths()


class JSONQueryProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(JSONQueryProcessor, self).__init__(envars, **kwargs)
        self.ext = ".json"
        self.content_type = "application/json"
        self.create_default_paths()


class JSONStreamQueryProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(JSONStreamQueryProcessor, self).__init__(envars, **kwargs)
        self.ext = ".json"
        self.content_type = "application/x-json-stream"
        self.create_default_paths()


class JSONEnvUpdateProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(JSONEnvUpdateProcessor, self).__init__(envars, **kwargs)
        self.query_keys = self.parameters.get("query_keys")

    def process(self):
        resp = self.catalogQuery(headers={'accept': "application/json"}, as_file=False)
        if resp:
            if isinstance(resp, list):
                resp = resp[0]
            if self.query_keys is not None:
                results = {key: resp[key] for key in self.query_keys}
            else:
                results = resp
            self.envars.update(results)
            self._urlencode_envars()
        return {}


class CreateDirProcessor(JSONEnvUpdateProcessor):
    def __init__(self, envars=None, **kwargs):
        super(CreateDirProcessor, self).__init__(envars, **kwargs)
        self.ext = ""

    def process(self):
        super(CreateDirProcessor, self).process()
        self.create_default_paths()
        make_dirs(self.output_abspath)




import json
import logging
from deriva.transfer.upload.processors import BaseProcessor

logger = logging.getLogger(__name__)


class MetadataUpdateProcessor(BaseProcessor):
    """
    Upload processor used to load external JSON into a dict for use as upload metadata.
    """
    def __init__(self, **kwargs):
        super(MetadataUpdateProcessor, self).__init__(**kwargs)
        self.metadata = kwargs.get("metadata", dict()) or dict

    def process(self):
        input_file = self.parameters.get("input_file")
        if not input_file:
            return

        with open(input_file, "r") as input_data:
            data = json.load(input_data)
            if not isinstance(data, dict):
                logger.warning("Type mismatch: expected dict object from loaded JSON file: %s" % input_file)
            for key in data.keys():
                if key in self.metadata:
                    logger.warning(
                        "Duplicate key '%s' encountered in metadata input file [%s] existing key will be "
                        "overwritten." % (key, input_file))
            self.metadata.update(data)



"""DERIVA command-line interface for validating annotations."""

import logging
import re
import sys
from deriva.core import BaseCLI, DerivaServer, tag, annotation, get_credential, format_credential

logger = logging.getLogger(__name__)

_epilog = """
Known tag names include: {tags}.
""".format(tags=', '.join(tag.values()))


class AnnotationValidateCLI (BaseCLI):
    """Annotation Validate Command-Line Interface"""

    def __init__(self):
        super(AnnotationValidateCLI, self).__init__(__doc__, _epilog, hostname_required=True)
        self.remove_options(['--config-file'])
        self.parser.add_argument('catalog', metavar='<catalog>', help="Catalog identifier.")
        self.parser.add_argument('-s', '--schema', metavar='<schema>', default='.*', help="Regular expression pattern for schema name")
        self.parser.add_argument('-t', '--table', metavar='<table>', default='.*', help="Regular expression pattern for table name")
        self.parser.add_argument('-c', '--column', metavar='<column>', default='.*', help="Regular expression pattern for column name")
        self.parser.add_argument('-k', '--key', metavar='<key>', default='.*', help="Regular expression pattern for key constraint name")
        self.parser.add_argument('-f', '--foreign-key', metavar='<foreign_key>', default='.*', help="Regular expression pattern for foreign key constraint name")
        self.parser.add_argument('-a', '--tag', metavar='<tag>', default=None, help="Tag name of annotation")
        self.parser.add_argument('--skip-model-names', action='store_true', help="Skip validation of model names found inside of annotations")

    def main(self):
        args = self.parse_cli()
        validate_model_names = not args.skip_model_names

        if args.token:
            credential = format_credential(token=args.token)
        elif args.oauth2_token:
            credential = format_credential(oauth2_token=args.oauth2_token)
        else:
            credential = get_credential(args.host, credential_file=args.credential_file)

        server = DerivaServer('https', args.host, credentials=credential)
        catalog = server.connect_ermrest(args.catalog)
        catalog.dcctx['cid'] = "cli/" + AnnotationValidateCLI.__name__
        model = catalog.getCatalogModel()
        errors = []
        num_objects_tested = 0

        # catalog annotation validation...
        logger.debug("Validating catalog annotations...")
        errors.extend(annotation.validate(model, tag_name=args.tag, validate_model_names=validate_model_names))
        num_objects_tested += 1

        for schema_name in model.schemas:
            if not re.search(args.schema, schema_name):
                continue

            # schema annotation validation...
            logger.debug("Validating '%s' annotations..." % schema_name)
            schema = model.schemas[schema_name]
            errors.extend(annotation.validate(schema, tag_name=args.tag, validate_model_names=validate_model_names))
            num_objects_tested += 1

            for table_name in schema.tables:
                if not re.search(args.table, table_name):
                    continue

                # table annotations validation
                logger.debug("Validating '%s:%s' annotations..." % (schema_name, table_name))
                table = model.schemas[schema_name].tables[table_name]
                errors.extend(annotation.validate(table, tag_name=args.tag, validate_model_names=validate_model_names))
                num_objects_tested += 1

                for column in table.column_definitions:
                    if not re.search(args.column, column.name):
                        continue

                    # column annotations validation
                    logger.debug("Validating '%s:%s:%s' annotations..." % (schema_name, table_name, column.name))
                    errors.extend(annotation.validate(column, tag_name=args.tag, validate_model_names=validate_model_names))
                    num_objects_tested += 1

                for key in table.keys:
                    if not re.search(args.key, key.constraint_name):
                        continue

                    # key annotations validation
                    logger.debug("Validating '%s:%s' annotations..." % (schema_name, key.constraint_name))
                    errors.extend(annotation.validate(key, tag_name=args.tag, validate_model_names=validate_model_names))
                    num_objects_tested += 1

                for fkey in table.foreign_keys:
                    if not re.search(args.foreign_key, fkey.constraint_name):
                        continue

                    # fkey annotations validation
                    logger.debug("Validating '%s:%s' annotations..." % (schema_name, fkey.constraint_name))
                    errors.extend(annotation.validate(fkey, tag_name=args.tag, validate_model_names=validate_model_names))
                    num_objects_tested += 1

        logger.info("Found %d error(s) in %d model object(s)." % (len(errors), num_objects_tested))
        return 1 if errors else 0


def main():
    cli = AnnotationValidateCLI()
    return cli.main()


if __name__ == '__main__':
    sys.exit(main())



import os
import datetime
import logging
from deriva.core import urlquote
from deriva.core.utils import mime_utils as mu, hash_utils as hu

PROCESSOR_PARAMS_KEY = "processor_params"
SERVICE_URL_KEY = "deriva_service_url"
SOURCE_URL_KEY = "source_url"
LOCAL_PATH_KEY = "local_path"
REMOTE_PATHS_KEY = "remote_paths"
FILE_SIZE_KEY = "size"
MD5_KEY = "md5"
SHA256_KEY = "sha256"
CONTENT_TYPE_KEY = "content_type"
IDENTIFIER_KEY = "identifier"
IDENTIFIER_LANDING_PAGE = "identifier_landing_page"

logger = logging.getLogger(__name__)


class BaseProcessor(object):
    """
    Base class for Processor classes
    """

    def __init__(self, envars=None, **kwargs):
        self.envars = envars if (envars is not None) else dict()
        self._urlencode_envars()
        self.kwargs = kwargs
        self.outputs = kwargs["inputs"]
        self.parameters = kwargs.get(PROCESSOR_PARAMS_KEY, dict()) or dict()
        self.identity = kwargs.get("identity", dict()) or dict()
        self.wallet = kwargs.get("wallet", dict()) or dict()
        self.timeout = kwargs.get("timeout", None)
        self.callback = self.parameters.get("callback", self.default_callback)
        self.last_timestamp = datetime.datetime.now()

    def _urlencode_envars(self, safe_overrides=None):
        urlencoded = dict()
        if not safe_overrides:
            safe_overrides = dict()
        for k, v in self.envars.items():
            if k.endswith("_urlencoded"):
                continue
            urlencoded[k + "_urlencoded"] = urlquote(str(v), safe_overrides.get(k, ""))
        self.envars.update(urlencoded)

    @classmethod
    def process(cls):
        raise NotImplementedError("Must be implemented by subclass")

    @staticmethod
    def create_paths(base_path, sub_path=None, filename=None, ext=None, is_bag=False, envars=None):
        relpath = sub_path if sub_path else ''
        if filename:
            relpath = ''.join([relpath, "/", filename]) if relpath else filename

        if isinstance(envars, dict):
            relpath = relpath.format(**envars)

        cur_ext = os.path.splitext(relpath)[1]
        if ext:
            ext = ext if ext.startswith(".") else "." + ext
            if not cur_ext == ext:
                relpath += ext

        abspath = os.path.abspath(
            os.path.join(base_path, 'data' if is_bag else '', relpath))

        return relpath, abspath

    @staticmethod
    def make_file_output_values(file_path, input_dict, make_file_hashes=True):
        input_dict[FILE_SIZE_KEY] = input_dict.get(FILE_SIZE_KEY, os.path.getsize(file_path))
        input_dict[CONTENT_TYPE_KEY] = input_dict.get(CONTENT_TYPE_KEY, mu.guess_content_type(file_path))
        has_file_hashes = input_dict.get(MD5_KEY) is not None and input_dict.get(SHA256_KEY) is not None
        if not has_file_hashes and make_file_hashes:
            input_dict.update(hu.compute_file_hashes(file_path, [MD5_KEY, SHA256_KEY]))

    def should_abort(self):
        try:
            if self.timeout and isinstance(self.timeout, datetime.datetime):
                now = datetime.datetime.now()
                elapsed = now - self.last_timestamp
                elapsed_time = str("Elapsed time since last timestamp: %s." % elapsed) if \
                    (elapsed > datetime.timedelta(milliseconds=0)) else ""
                if elapsed_time:
                    logging.debug(elapsed_time)
                if now > self.timeout:
                    return True
            return False
        finally:
            self.last_timestamp = datetime.datetime.now()

    def default_callback(self, **kwargs):
        progress = kwargs.get("progress", None)
        if progress and logger.isEnabledFor(logging.DEBUG):
            logger.debug(progress)
        if self.should_abort():
            return False
        return True



import os
import sys
import copy
import time
import logging
import datetime
from deriva.core import get_credential, urlquote
from deriva.transfer import DerivaDownload
from deriva.transfer.backup import DerivaBackupError, DerivaBackupConfigurationError, \
    DerivaBackupAuthenticationError, DerivaBackupAuthorizationError


class DerivaBackup(DerivaDownload):
    BASE_CONFIG = {
        "catalog": {
            "query_processors": []
        }
    }
    BASE_SCHEMA_QUERY_PROC = {
        "processor": "json",
        "processor_params": {
            "query_path": "/schema",
            "output_path": "catalog-schema"
        }
    }
    BASE_DATA_QUERY_PATH = "/entity/{}:{}"
    BASE_DATA_OUTPUT_PATH = "records/{}/{}"
    BASE_ASSET_OUTPUT_PATH = "assets"

    def __init__(self, *args, **kwargs):
        DerivaDownload.__init__(self, *args, **kwargs)

        self.config_file = kwargs.get("config_file")
        self.annotation_config = None

        if not self.config:
            self.config = copy.deepcopy(self.BASE_CONFIG)

            no_schema = kwargs.get("no_schema", False)
            if not no_schema:
                self.config["catalog"]["query_processors"].append(self.BASE_SCHEMA_QUERY_PROC)

            no_bag = kwargs.get("no_bag", False)
            if not no_bag:
                bag = dict()
                bag["bag_name"] = os.path.basename(self.output_dir)
                bag["bag_archiver"] = kwargs.get("bag_archiver", "tgz")
                bag["bag_algorithms"] = ["sha256", "md5"]
                self.config["bag"] = bag

            # if credentials have not been explicitly set yet, try to get them from the default credential store
            if not self.credentials:
                self.set_credentials(get_credential(self.hostname))

            logging.debug("Inspecting catalog model...")
            model = self.catalog.getCatalogModel()
            # if we dont have catalog ownership rights, its a hard error for now
            if not model.acls:
                raise DerivaBackupAuthorizationError("Only catalog owners may perform full catalog dumps.")

            if kwargs.get("no_data", False):
                return

            exclude = kwargs.get("exclude_data", list())
            for sname, schema in model.schemas.items():
                if sname in exclude:
                    logging.info("Excluding data dump from all tables in schema: %s" % sname)
                    continue
                for tname, table in schema.tables.items():
                    fqtname = "%s:%s" % (sname, tname)
                    if table.kind != "table":
                        logging.warning("Skipping data dump of %s: %s" % (table.kind, fqtname))
                        continue
                    if fqtname in exclude:
                        logging.info("Excluding data dump from table: %s" % fqtname)
                        continue
                    if "RID" not in table.column_definitions.elements:
                        logging.warning(
                            "Source table %s.%s lacks system-columns and will not be dumped." % (sname, tname))
                        continue

                    # Configure table data download query processors
                    data_format = "json" if (sname, tname) in {
                        ('public', 'ERMrest_Client'),
                        ('public', 'ERMrest_Group'),
                    } else "json-stream"
                    q_sname = urlquote(sname)
                    q_tname = urlquote(tname)
                    output_path = self.BASE_DATA_OUTPUT_PATH.format(q_sname, q_tname)
                    query_path = self.BASE_DATA_QUERY_PATH.format(q_sname, q_tname)
                    query_proc = dict()
                    query_proc["processor"] = data_format
                    query_proc_params = {"query_path": query_path, "output_path": output_path}
                    if data_format in ("json-stream", "csv"):
                        query_proc_params.update({"paged_query": True, "paged_query_size": 100000})
                    query_proc["processor_params"] = query_proc_params
                    self.config["catalog"]["query_processors"].append(query_proc)

        self.generate_asset_configs()

    def generate_asset_configs(self):
        # TODO: Generate asset data download query processor configuration entries
        pass

    def download(self, **kwargs):
        logging.info("Backing up catalog: %s" % self.catalog.get_server_uri())
        success = True
        start = datetime.datetime.now()
        try:
            return super(DerivaBackup, self).download(**kwargs)
        except:
            success = False
            raise
        finally:
            elapsed_time = datetime.datetime.now() - start
            total_secs = elapsed_time.total_seconds()
            elapsed = time.strftime('%H:%M:%S', time.gmtime(total_secs))
            logging.info("Backup of catalog %s %s. %s" % (self.catalog.get_server_uri(),
                                                          "completed successfully" if success else "failed",
                                                          ("Elapsed time: %s" % elapsed) if (total_secs > 0) else ""))



import os
import json
import uuid
import datetime
import logging
import requests
import certifi
from bdbag import bdbag_ro as ro
from deriva.core import urlsplit, format_exception, get_transfer_summary, make_dirs, DEFAULT_CHUNK_SIZE
from deriva.core.utils.mime_utils import parse_content_disposition
from deriva.transfer.download.processors.query.base_query_processor import BaseQueryProcessor, \
    LOCAL_PATH_KEY, FILE_SIZE_KEY
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
    DerivaDownloadAuthenticationError


class FileDownloadQueryProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(FileDownloadQueryProcessor, self).__init__(envars, **kwargs)
        self.content_type = "application/x-json-stream"
        filename = ''.join(['download-manifest_', str(uuid.uuid4()), ".json"])
        self.output_relpath, self.output_abspath = self.create_paths(self.base_path, filename=filename)
        self.ro_file_provenance = False
        self.allow_anonymous = kwargs.get("allow_anonymous", True)

    def process(self):
        if not self.identity and not self.allow_anonymous:
            raise DerivaDownloadAuthenticationError(
                "Unauthenticated (anonymous) users are not permitted to request direct file downloads.")
        super(FileDownloadQueryProcessor, self).process()
        self.outputs.update(self.downloadFiles(self.output_abspath))
        return self.outputs

    def getExternalFile(self, url, output_path, headers=None):
        host = urlsplit(url).netloc
        if output_path:
            if not headers:
                headers = self.HEADERS.copy()
            else:
                headers.update(self.HEADERS)
            session = self.getExternalSession(host)
            with session.get(url, headers=headers, stream=True, verify=certifi.where()) as r:
                if r.status_code != 200:
                    file_error = "File [%s] transfer failed." % output_path
                    url_error = 'HTTP GET Failed for url: %s' % url
                    host_error = "Host %s responded:\n\n%s" % (urlsplit(url).netloc, r.text)
                    raise DerivaDownloadError('%s\n\n%s\n%s' % (file_error, url_error, host_error))
                else:
                    total = 0
                    start = datetime.datetime.now()
                    logging.debug("Transferring file %s to %s" % (url, output_path))
                    with open(output_path, 'wb') as data_file:
                        for chunk in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
                            data_file.write(chunk)
                            total += len(chunk)
                    elapsed = datetime.datetime.now() - start
                    summary = get_transfer_summary(total, elapsed)
                    logging.info("File [%s] transfer successful. %s" % (output_path, summary))
                    length = int(r.headers.get('Content-Length'))
                    content_type = r.headers.get("Content-Type")
                    return output_path, length, content_type

    def downloadFiles(self, input_manifest):
        logging.info("Attempting to download file(s) based on the results of query: %s" % self.query)
        try:
            with open(input_manifest, "r", encoding='utf-8') as in_file:
                file_list = dict()
                for line in in_file:
                    entry = json.loads(line)
                    url = entry.get('url')
                    if not url:
                        logging.warning(
                            "Skipping download due to missing required attribute \"url\" in download manifest entry %s"
                            % json.dumps(entry))
                        continue
                    store = self.getHatracStore(url)
                    filename = entry.get('filename') if not self.output_filename else self.output_filename
                    if not filename:
                        if store:
                            try:
                                head = store.head(url, headers=self.HEADERS)
                            except requests.HTTPError as e:
                                raise DerivaDownloadError("HEAD request for [%s] failed: %s" % (url, e))
                            content_disposition = head.headers.get("Content-Disposition") if head.ok else None
                            filename = os.path.basename(filename).split(":")[0] if not content_disposition else \
                                parse_content_disposition(content_disposition)
                        else:
                            filename = os.path.basename(url)
                    env = self.envars.copy()
                    env.update(entry)
                    rel_path, file_path = self.create_paths(self.base_path,
                                                            sub_path=self.sub_path,
                                                            filename=filename,
                                                            is_bag=self.is_bag,
                                                            envars=env)
                    output_dir = os.path.dirname(file_path)
                    make_dirs(output_dir)
                    if store:
                        try:
                            resp = store.get_obj(url, self.HEADERS, file_path)
                        except requests.HTTPError as e:
                            raise DerivaDownloadError("File [%s] transfer failed: %s" % (file_path, e))
                        length = int(resp.headers.get('Content-Length'))
                        content_type = resp.headers.get("Content-Type")
                        url = self.getExternalUrl(url)
                    else:
                        url = self.getExternalUrl(url)
                        file_path, length, content_type = self.getExternalFile(url, file_path)
                    file_bytes = os.path.getsize(file_path)
                    if length != file_bytes:
                        raise DerivaDownloadError(
                            "File size of %s does not match expected size of %s for file %s" %
                            (length, file_bytes, file_path))
                    if self.ro_manifest:
                        ro.add_file_metadata(self.ro_manifest,
                                             source_url=url,
                                             local_path=rel_path,
                                             media_type=content_type,
                                             retrieved_on=ro.make_retrieved_on(),
                                             retrieved_by=ro.make_retrieved_by(
                                                 self.ro_author_name, orcid=self.ro_author_orcid),
                                             bundled_as=ro.make_bundled_as())
                    file_list.update({rel_path: {LOCAL_PATH_KEY: file_path, FILE_SIZE_KEY: file_bytes}})
                    if self.callback:
                        if not self.callback(progress="Downloaded [%s] to: %s" % (url, file_path)):
                            break

                return file_list
        finally:
            os.remove(input_manifest)



from deriva.core import BaseCLI, __version__
from deriva.seo import SitemapBuilder
import sys
import json
import logging

def main():
    logger = logging.getLogger(__name__)
    logger.setLevel("WARNING")
    cli = BaseCLI(__name__,
                  "Create a sitemap from a table specified on the command line or a set of tables from a config file",
                  __version__, hostname_required=True)
    cli.remove_options(["--credential-file", "--token", "--oauth2-token"])
    cli.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number")
    cli.parser.add_argument("-p", "--priority", type=float,
                            help="A floating-point number between 0.0 and 1.0 indicating the table's priority")
    cli.parser.add_argument("-s", "--schema", help="the name of the schema of the (single) table to include")
    cli.parser.add_argument("-t", "--table", help="the name of the (single) table to include")
    args = cli.parse_cli()

    if args.priority is not None:
        if args.priority < 0 or args.priority > 1:
            logger.error("priority should be a floating-point number between 0 and 1")
            sys.exit(1)

    if not ((args.schema and args.table) or args.config_file):
        logger.error("must specify either a schema and table or a config file")
        sys.exit(1)

    sb = SitemapBuilder("https", args.host, args.catalog)
    if args.schema and args.table:
        sb.add_table_spec(args.schema, args.table, priority=args.priority)
    if args.config_file:
        rows = json.load(open(args.config_file))
        for row in rows:
            if row.get("schema") is None or row.get("table") is None:
                logger.warning("malformed entry in {f}: schema or table is missing. Skipping".format(f=args.config_file))
                next
            if row.get("priority") is None:
                priority = args.priority
            else:
                try:
                    priority = float(row.get("priority"))
                    if priority < 0 or priority > 1:
                        logger.warning("bad priority '{p}' - should be a floating-point number between 0 and 1. Ignoring".format(p=priority))
                        priority = args.priority
                except ValueError:
                    logger.warning("malformed priority '{p}' - should be a floating-point number between 0 and 1. Ignoring".format(p=row.get("priority")))
                    priority = args.priority

            sb.add_table_spec(row["schema"], row["table"], priority=priority)
    sb.write_sitemap(sys.stdout)
    return 0

if __name__ == '__main__':
    sys.exit(main())



from .deriva_binding import DerivaBinding
from .ermrest_catalog import ErmrestCatalog, ErmrestSnapshot, ErmrestAlias, DerivaServer


# DerivaServer class definition moved to avoid circular imports...



import sys
from importlib import import_module
from deriva.core import read_config, format_exception
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import BaseProcessor
from deriva.transfer.download.processors.query.base_query_processor import CSVQueryProcessor, \
    JSONEnvUpdateProcessor, JSONQueryProcessor, JSONStreamQueryProcessor, CreateDirProcessor
from deriva.transfer.download.processors.query.bag_fetch_query_processor import BagFetchQueryProcessor
from deriva.transfer.download.processors.query.file_download_query_processor import FileDownloadQueryProcessor
from deriva.transfer.download.processors.transform.fasta_transform_processor import FastaExportTransformProcessor
from deriva.transfer.download.processors.transform.geo_transform_processor import GeoExportTransformProcessor
from deriva.transfer.download.processors.transform.string_transform_processor import InterpolationTransformProcessor, \
    StrSubTransformProcessor, ConcatenateTransformProcessor
from deriva.transfer.download.processors.transform.format_transform_processor import JSONtoCSVTransformProcessor
from deriva.transfer.download.processors.transform.column_transform_processor import ColumnTransformProcessor
from deriva.transfer.download.processors.postprocess.identifier_post_processor import FAIRIdentifierPostProcessor
from deriva.transfer.download.processors.postprocess.transfer_post_processor import Boto3UploadPostProcessor
from deriva.transfer.download.processors.postprocess.url_post_processor import UrlRewritePostProcessor

DEFAULT_QUERY_PROCESSORS = {
    "csv": CSVQueryProcessor,
    "env": JSONEnvUpdateProcessor,
    "dir": CreateDirProcessor,
    "json": JSONQueryProcessor,
    "json-stream": JSONStreamQueryProcessor,
    "download": FileDownloadQueryProcessor,
    "fetch": BagFetchQueryProcessor
}

DEFAULT_TRANSFORM_PROCESSORS = {
    "cat": ConcatenateTransformProcessor,
    "interpolation": InterpolationTransformProcessor,
    "strsub": StrSubTransformProcessor,
    "geo": GeoExportTransformProcessor,
    "fasta": FastaExportTransformProcessor,
    "json2csv": JSONtoCSVTransformProcessor,
    "column": ColumnTransformProcessor
}

DEFAULT_POST_PROCESSORS = {
    "identifier": FAIRIdentifierPostProcessor,
    "cloud_upload": Boto3UploadPostProcessor,
    "url_rewrite": UrlRewritePostProcessor
}


def find_processor(processor_name, processor_type=None, defaults={}, **kwargs):
    if not processor_type:
        if processor_name in defaults:
            return defaults[processor_name]
        else:
            raise DerivaDownloadConfigurationError("Unsupported processor type: %s" % processor_name)

    if not is_processor_whitelisted(processor_type, **kwargs):
        raise DerivaDownloadConfigurationError(
            "Unknown external processor type [%s]: this processor must be added to the whitelist." % processor_type)

    clazz = None
    try:
        module_name, class_name = processor_type.rsplit(".", 1)
        try:
            module = sys.modules[module_name]
        except KeyError:
            module = import_module(module_name)
        clazz = getattr(module, class_name) if module else None
    except (ImportError, AttributeError):
        pass
    if not clazz:
        raise DerivaDownloadConfigurationError("Unable to import specified processor class: %s" % processor_type)

    if not issubclass(clazz, BaseProcessor):
        raise DerivaDownloadError(format_exception(NotImplementedError("The imported class %s is not a subclass of %s" %
                                  (processor_type, BaseProcessor.__module__ + "." + BaseProcessor.__name__))))

    return clazz


def is_processor_whitelisted(processor_type, **kwargs):
    config_file_path = kwargs.get("config_file_path")
    config = read_config(config_file=config_file_path)
    whitelist = config.get("download_processor_whitelist", [])

    return processor_type in whitelist


def find_query_processor(processor_name, processor_type=None, **kwargs):
    return find_processor(processor_name, processor_type, DEFAULT_QUERY_PROCESSORS, **kwargs)


def find_transform_processor(processor_name, processor_type=None, **kwargs):
    return find_processor(processor_name, processor_type, DEFAULT_TRANSFORM_PROCESSORS, **kwargs)


def find_post_processor(processor_name, processor_type=None, **kwargs):
    return find_processor(processor_name, processor_type, DEFAULT_POST_PROCESSORS, **kwargs)



import platform
from deriva.core import get_credential, BaseCLI, DerivaServer, __version__


def rollback_annotation(host, catalog_id, snaptime=None, prot='https', credential=None):
    """
    Rollback the entire annotation hierarchy for the specified catalog to a given point in time specified by snaptime.
    """
    if not credential:
        credential = get_credential(host)
    server = DerivaServer(prot, host, credentials=credential)

    good_catalog = server.connect_ermrest(catalog_id, snaptime)
    good_config = good_catalog.getCatalogModel()

    live_catalog = server.connect_ermrest(catalog_id)
    live_config = live_catalog.getCatalogModel()

    # copy over annotations

    # catalog-level
    live_config.annotations.clear()
    live_config.annotations.update(good_config.annotations)

    for sname, live_schema in live_config.schemas.items():
        good_schema = good_config.schemas[sname]

        # schema-level
        live_schema.annotations.clear()
        live_schema.annotations.update(good_schema.annotations)

        for tname, live_table in live_schema.tables.items():
            if tname not in good_schema.tables:
                print('Warning: skipping live table %s.%s which lacks known-good table' % (sname, tname))
                continue

            good_table = good_schema.tables[tname]

            # table-level
            live_table.annotations.clear()
            live_table.annotations.update(good_table.annotations)

            for live_column in live_table.column_definitions:
                cname = live_column.name
                good_column = good_table.column_definitions[cname]

                # column-level
                live_column.annotations.clear()
                live_column.annotations.update(good_column.annotations)

            for live_key in live_table.keys:
                constr_name = tuple(live_key.names[0])
                good_key = live_table.keys[constr_name]

                # key-level
                live_key.annotations.clear()
                live_key.annotations.update(good_key.annotations)

            for live_fkey in live_table.foreign_keys:
                constr_name = tuple(live_fkey.names[0])
                good_fkey = live_table.foreign_keys[constr_name]

                # fkey-level
                live_fkey.annotations.clear()
                live_fkey.annotations.update(good_fkey.annotations)

    live_config.apply(live_catalog)


def main():
    cli = BaseCLI("annotation rollback tool", None, version=__version__, hostname_required=True)
    cli.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number. Default: 1")
    cli.parser.add_argument("--snapshot", metavar="<snapshot ID", help="Catalog snapshot ID. Example: 2QG-VWP6-0YG0")
    args = cli.parse_cli()
    credential = get_credential(args.host, args.credential_file)
    rollback_annotation(args.host, args.catalog, snaptime=args.snapshot, credential=credential)


if __name__ == '__main__':
    sys.exit(main())



import os
import pathlib
import logging
from deriva.core import stob
from deriva.transfer.upload.processors.base_processor import BaseProcessor, PROCESSOR_MODIFIED_FILE_PATH_KEY
from bdbag import bdbag_api as bdb

logger = logging.getLogger(__name__)


class ArchiveProcessor(BaseProcessor):
    """
    """
    def __init__(self, **kwargs):
        super(ArchiveProcessor, self).__init__(**kwargs)


class BagArchiveProcessor(ArchiveProcessor):
    """
    Upload preprocessor that takes an input directory and creates a BDBag archive from it.
    """
    def __init__(self, **kwargs):
        super(BagArchiveProcessor, self).__init__(**kwargs)
        self.file_path = kwargs.get("file_path")
        self.metadata = kwargs.get("metadata")
        assert self.file_path is not None
        self.processor_params = kwargs.get("processor_params") or dict()

    def process(self):
        processor_output = self.kwargs.get("processor_output")
        idempotent_archive = stob(self.processor_params.get("idempotent_archive", True))
        bdb.make_bag(self.file_path, update=True, idempotent=idempotent_archive)
        try:
            archive_file = (
                bdb.archive_bag(self.file_path,
                                self.processor_params.get("format", "zip"),
                                idempotent=idempotent_archive))
            self.metadata["file_name"] = os.path.basename(archive_file)
            self.metadata["file_size"] = os.path.getsize(archive_file)
            self.metadata["base_path"] = os.path.dirname(archive_file)
            self.metadata["file_ext"] = "".join(pathlib.PurePath(archive_file).suffixes)
            self.metadata["base_name"] = self.metadata["file_name"].rsplit(self.metadata["file_ext"])[0]
            if processor_output is not None:
                processor_output.update({PROCESSOR_MODIFIED_FILE_PATH_KEY: archive_file})
        except:
            raise
        finally:
            bdb.revert_bag(self.file_path)
        return processor_output




#
# Copyright 2017 University of Southern California
# Distributed under the Apache License, Version 2.0. See LICENSE for more info.
#

""" Installation script for the deriva package.
"""

from setuptools import setup, find_packages
import re
import io

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
    io.open('deriva/core/__init__.py', encoding='utf_8_sig').read()
    ).group(1)


def get_readme_contents():
    with io.open('README.md') as readme_file:
        return readme_file.read()


url = "https://github.com/informatics-isi-edu/deriva-py"
author = 'USC Information Sciences Institute, Informatics Systems Research Division'
author_email = 'isrd-support@isi.edu'


setup(
    name='deriva',
    description='Python APIs and CLIs (Command-Line Interfaces) for the DERIVA platform.',
    long_description='For further information, visit the project [homepage](%s).' % url,
    long_description_content_type='text/markdown',
    url=url,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    version=__version__,
    packages=find_packages(exclude=["tests"]),
    package_data={
        'deriva.config': ['examples/*.json'],
        'deriva.core': ['schemas/*.schema.json']
    },
    python_requires='>=3.8, <4',
    entry_points={
        'console_scripts': [
            'deriva-upload-cli = deriva.transfer.upload.__main__:main',
            'deriva-download-cli = deriva.transfer.download.__main__:main',
            'deriva-catalog-cli = deriva.core.catalog_cli:main',
            'deriva-hatrac-cli = deriva.core.hatrac_cli:main',
            'deriva-acl-config = deriva.config.acl_config:main',
            'deriva-annotation-config = deriva.config.annotation_config:main',
            'deriva-annotation-dump = deriva.config.dump_catalog_annotations:main',
            'deriva-annotation-rollback = deriva.config.rollback_annotation:main',
            'deriva-annotation-validate = deriva.config.annotation_validate:main',
            'deriva-sitemap-cli = deriva.seo.sitemap_cli:main',
            'deriva-backup-cli = deriva.transfer.backup.__main__:main',
            'deriva-restore-cli = deriva.transfer.restore.__main__:main',
            'deriva-globus-auth-utils = deriva.core.utils.globus_auth_utils:main'
        ]
    },
    install_requires=[
        'packaging',
        'requests',
        'certifi',
        'pika',
        'urllib3>=1.26,<3',
        'portalocker>=1.2.1',
        'bdbag>=1.7.2',
        'globus_sdk>=3,<4',
        'fair-research-login>=0.3.1',
        'fair-identifiers-client>=0.5.1',
        'jsonschema>=3.1'
    ],
    license='Apache 2.0',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ]
)



import os
from deriva.core import stob
from deriva.core.utils.mime_utils import guess_content_type
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import BaseProcessor, \
    LOCAL_PATH_KEY, FILE_SIZE_KEY, SOURCE_URL_KEY
from bdbag import bdbag_ro as ro


class BaseTransformProcessor(BaseProcessor):
    """
    Base class for TransformProcessor classes
    """

    def __init__(self, envars=None, **kwargs):
        super(BaseTransformProcessor, self).__init__(envars, **kwargs)
        self.base_path = kwargs["base_path"]
        self.input_paths = self.parameters.get("input_paths", [])
        if not self.input_paths:
            self.input_paths = [self.parameters["input_path"]]  # for backward compatibility
        self.sub_path = self.parameters.get("output_path")
        self.output_filename = self.parameters.get("output_filename")
        self.is_bag = kwargs.get("bag", False)
        self.transformed_output = self.outputs.get(self.input_path, dict())
        self.url = self.transformed_output.get(SOURCE_URL_KEY)
        self.ro_file_provenance = stob(self.parameters.get("ro_file_provenance", False if not self.is_bag else True))
        self.ro_manifest = self.kwargs.get("ro_manifest")
        self.ro_author_name = self.kwargs.get("ro_author_name")
        self.ro_author_orcid = self.kwargs.get("ro_author_orcid")
        self.delete_input = stob(self.parameters.get("delete_input", True))
        self.input_relpaths = []
        self.input_abspaths = []
        self.output_relpath = None
        self.output_abspath = None

    @property
    def input_path(self):  # for backward compatibility
        return self.input_paths[0]

    @property
    def input_relpath(self):  # for backward compatibility
        return self.input_relpaths[0]

    @property
    def input_abspath(self):  # for backward compatibility
        return self.input_abspaths[0]

    def _create_input_output_paths(self):
        for input_path in self.input_paths:
            relpath, abspath = self.create_paths(self.base_path,
                                                 sub_path=input_path,
                                                 is_bag=self.is_bag,
                                                 envars=self.envars)
            self.input_relpaths.append(relpath)
            self.input_abspaths.append(abspath)
        self.output_relpath, self.output_abspath = self.create_paths(self.base_path,
                                                                     sub_path=self.sub_path,
                                                                     filename=self.output_filename,
                                                                     is_bag=self.is_bag,
                                                                     envars=self.envars)

    def _delete_input(self):
        for input_abspath in self.input_abspaths:
            if os.path.isfile(input_abspath):
                os.remove(input_abspath)
        for input_relpath in self.input_relpaths:
            del self.outputs[input_relpath]

    def process(self):
        if self.ro_manifest and self.ro_file_provenance:
            ro.add_file_metadata(self.ro_manifest,
                                 source_url=self.url,
                                 local_path=self.output_relpath,
                                 media_type=guess_content_type(self.output_abspath),
                                 retrieved_on=ro.make_retrieved_on(),
                                 retrieved_by=ro.make_retrieved_by(self.ro_author_name, orcid=self.ro_author_orcid),
                                 bundled_as=ro.make_bundled_as())
        if self.delete_input:
            self._delete_input()

        self.outputs.update({self.output_relpath: {LOCAL_PATH_KEY: self.output_abspath,
                                                   FILE_SIZE_KEY: os.path.getsize(self.output_abspath),
                                                   SOURCE_URL_KEY: self.url}})
        return self.outputs






import json
import logging
from importlib import import_module
from deriva.core import get_credential, urlsplit, urlunsplit, format_exception, stob
from deriva.core.utils.globus_auth_utils import GlobusNativeLogin
from deriva.core.utils.webauthn_utils import get_wallet_entries
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import *
from fair_research_login.client import NoSavedTokens, TokensExpired
from fair_identifiers_client.identifiers_api import IdentifierClient, AccessTokenAuthorizer, IdentifierClientError


class FAIRIdentifierPostProcessor(BaseProcessor):
    """
    Post processor that mints identifiers for download results using FAIR-Research Identifier Client.
    """

    IDENTIFIER_SERVICE = "https://identifiers.fair-research.org/"
    IDENTIFIER_SERVICE_TEST = "https://identifiers-test.fair-research.org/"
    IDENTIFIER_SERVICE_WRITER_SCOPE = "https://auth.globus.org/scopes/identifiers.fair-research.org/writer"
    TEST_IDENTIFIER_NAMESPACE = "minid-test"
    IDENTIFIER_NAMESPACE = "minid"

    def __init__(self, envars=None, **kwargs):
        super(FAIRIdentifierPostProcessor, self).__init__(envars, **kwargs)

    def load_identifier_client(self, identifiers_service_url):
        if not self.identity:
            logging.warning("Unauthenticated (anonymous) identity being used with identifier client")
        if self.wallet:
            entries = get_wallet_entries(self.wallet, "oauth2",
                                         credential_source="https://auth.globus.org",
                                         scopes=[self.IDENTIFIER_SERVICE_WRITER_SCOPE])
            token = entries[0].get("access_token") if entries else None
        else:
            tokens = None
            host = self.envars["hostname"]
            gnl = GlobusNativeLogin(host=host)
            try:
                tokens = gnl.client.load_tokens()
            except NoSavedTokens:
                pass
            except TokensExpired as e:
                raise RuntimeError(
                    "Unable to obtain token set due to refresh token expiry. Please logout of the expired scopes. %s" %
                    format_exception(e))
            if not tokens:
                raise RuntimeError("Login required. No saved tokens.")
            token = gnl.find_access_token_for_scope(self.IDENTIFIER_SERVICE_WRITER_SCOPE, tokens)
        ac = AccessTokenAuthorizer(token) if token else None
        return IdentifierClient(base_url=identifiers_service_url, app_name="DERIVA Export", authorizer=ac)

    def process(self):
        test = stob(self.parameters.get("test", "False"))
        test_service = stob(self.parameters.get("test_service", "False"))
        identifiers_service_url = self.IDENTIFIER_SERVICE_TEST if test_service else self.IDENTIFIER_SERVICE
        ic = self.load_identifier_client(identifiers_service_url)
        namespace = (self.TEST_IDENTIFIER_NAMESPACE if test else self.IDENTIFIER_NAMESPACE)
        for k, v in self.outputs.items():
            file_path = v[LOCAL_PATH_KEY]
            self.make_file_output_values(file_path, v)
            checksum = v[SHA256_KEY][0]
            title = self.parameters.get("title", "%s" % k)
            metadata = {"title": title}
            length = v[FILE_SIZE_KEY]
            if length is not None:
                metadata.update({"length": length})
            author = self.parameters.get("author")
            if author and stob(self.parameters.get("include_author", "True")):
                metadata.update({"author": author})
            created_by = self.identity.get('full_name') if self.identity else None
            if created_by:
                metadata.update({"created_by": created_by})
            visible_to = self.parameters.get("visible_to", ["public"])
            locations = v.get(REMOTE_PATHS_KEY) or self.parameters.get("locations")
            if not locations:
                raise DerivaDownloadConfigurationError(
                    "Invalid URLs: One or more location URLs must be specified when registering an identifier.")
            env_column_map = self.parameters.get("env_column_map")
            if env_column_map:
                env_metadata = {key: val.format(**self.envars) for key, val in env_column_map.items()}
                metadata.update(env_metadata)
            kwargs = {
                "namespace": namespace,
                "visible_to": visible_to,
                "location": locations,
                "checksums": [{
                    "function": "sha256",
                    "value": checksum
                }],
                "metadata": metadata
            }
            try:
                logging.info("Attempting to create identifier for file [%s] with locations: %s" %
                             (file_path, locations))
                minid = ic.create_identifier(**kwargs)
                identifier = minid["identifier"]
                v[IDENTIFIER_KEY] = identifier
                v[IDENTIFIER_LANDING_PAGE] = \
                    self.parameters.get("redirect_base", "") + identifiers_service_url + identifier
            except IdentifierClientError as e:
                raise DerivaDownloadError("Unable to create identifier: %s" % format_exception(e))

        return self.outputs



import os
import shutil
import pathlib
import logging
import re
from deriva.transfer.upload.processors.base_processor import BaseProcessor, PROCESSOR_MODIFIED_FILE_PATH_KEY

logger = logging.getLogger(__name__)


class RenameProcessor(BaseProcessor):
    """
    """
    def __init__(self, **kwargs):
        super(RenameProcessor, self).__init__(**kwargs)


class FileRenameProcessor(RenameProcessor):
    """
    Upload preprocessor that renames an input file before uploading it. Note that by default this processor does not
    modify the actual input file name on disk, but rather the associated file path metadata that is propagated to the
    catalog and object store. If physical file renaming is desired, the processor param "rename_physical_file" can be
    set to True. Depending on how this processor is configured, a physically renamed file may no longer match the scan
    criteria of the upload configuration if the upload process is run again.
    """
    def __init__(self, **kwargs):
        super(FileRenameProcessor, self).__init__(**kwargs)
        self.file_path = kwargs.get("file_path")
        self.metadata = kwargs.get("metadata")
        assert self.file_path is not None
        self.processor_params = kwargs.get("processor_params") or dict()

    def process(self):
        path_dict = dict()
        processor_output = self.kwargs.get("processor_output")
        pattern = self.processor_params.get("pattern")
        if pattern:
            path_dict.update(re.match(pattern, self.file_path).groupdict() or dict())
        repl = self.processor_params.get("replacement")
        path_dict.update(self.metadata)
        file_path = os.path.realpath(repl.format(**path_dict))
        self.metadata["file_name"] = os.path.basename(file_path)
        self.metadata["file_ext"] = "".join(pathlib.PurePath(file_path).suffixes)
        self.metadata["base_path"] = os.path.dirname(file_path)
        self.metadata["base_name"] = self.metadata["file_name"].rsplit(self.metadata["file_ext"])[0]
        if file_path != self.file_path and self.processor_params.get("rename_physical_file", False):
            shutil.move(self.file_path, file_path)
            logger.info("Renamed input file %s to: %s" % (self.file_path, file_path))
            if processor_output is not None:
                processor_output.update({PROCESSOR_MODIFIED_FILE_PATH_KEY: file_path})

        return processor_output

